------------------------------------------------------
-- HuntersHelper.lua
-- Version 1700.3
------------------------------------------------------

-- Saved configuration & info
FHH_Config = { };
FHH_Config.Tooltip = true;

FHH_AbilityInfo = { };
-- Has the following internal structure:
--		REALM_PLAYER = {
--			SKILL
--		}

-- Runtime state
FHH_State = { };
FHH_State.RealmPlayer = nil;
FHH_State.TamingCritter = nil;

-- Constants
ALL_FAMILIES = "all beast families";
PET_TRAINER = "Pet Trainers (found in major cities and some towns)";
MAX_REPORTED_ZONES = 4;

function FHH_OnLoad()

	this:RegisterEvent("PLAYER_ENTERING_WORLD");
	this:RegisterEvent("UPDATE_MOUSEOVER_UNIT");

	-- Register Slash Commands
	SLASH_FHH1 = "/huntershelper";
	SLASH_FHH2 = "/hh";
	SlashCmdList["FHH"] = function(msg)
		FHH_ChatCommandHandler(msg);
	end
	
	GFWUtils.Print("Fizzwidget Hunter's Helper initialized!");
	
end

function FHH_OnEvent(event)

	if ( event == "PLAYER_ENTERING_WORLD" ) then
		
		_, realClass = UnitClass("player");
		if (realClass == "HUNTER") then
			-- only do stuff related to taming and checking hunter spells if you're a hunter.
			this:RegisterEvent("UNIT_AURA");
			this:RegisterEvent("UNIT_NAME_UPDATE");
			this:RegisterEvent("CRAFT_SHOW");
			this:RegisterEvent("CHAT_MSG_SYSTEM");

			if (FHH_State.RealmPlayer == nil) then
				FHH_State.RealmPlayer = GetCVar("realmName") .. "." .. UnitName("player");
			end
			if (FHH_AbilityInfo[FHH_State.RealmPlayer] == nil or table.getn(FHH_AbilityInfo[FHH_State.RealmPlayer]) == 0) then
				
				local _, myClass = UnitClass("player");
				if (myClass == "HUNTER" and UnitLevel("player") > 9) then
					FHH_AbilityInfo[FHH_State.RealmPlayer] = { };
					GFWUtils.Print("Hunter's Helper needs to collect info about what pet skills you already know; please open your Beast Training window. (Info on future skills will be collected as they are learned.)");
				end
			end
		end
		
	elseif ( event == "UPDATE_MOUSEOVER_UNIT" ) then
	
		if ( UnitExists("mouseover") and not UnitPlayerControlled("mouseover") and FHH_Config.Tooltip ) then

			local _, myClass = UnitClass("player");
			if (FHH_Config.Tooltip == "hunter" and myClass ~= "HUNTER") then return; end
			
			local creepName = UnitName("mouseover");
			local abilitiesList = nil;
			
			if (FHH_NewBeastInfo ~= nil and FHH_NewBeastInfo[creepName] ~= nil) then
				abilitiesList = GFWTable.Copy(FHH_NewBeastInfo[creepName]);
			elseif (FHH_BeastInfo[creepName] ~= nil) then
				abilitiesList = GFWTable.Copy(FHH_BeastInfo[creepName]);
				if (FHH_BadBeastInfo ~= nil and FHH_BadBeastInfo[creepName] ~= nil) then
					abilitiesList = GFWTable.Subtract(abilitiesList, FHH_BadBeastInfo[creepName]);
				end
			end
						
			if (abilitiesList) then
			
				local _, myClass = UnitClass("player");
				if (myClass == "HUNTER" and FHH_AbilityInfo[FHH_State.RealmPlayer] ~= nil and table.getn(FHH_AbilityInfo[FHH_State.RealmPlayer]) > 0) then
					for i=1, table.getn(abilitiesList) do
						if (GFWTable.IndexOf(FHH_AbilityInfo[FHH_State.RealmPlayer], abilitiesList[i]) ~= 0) then
							abilitiesList[i] = GRAY_FONT_COLOR_CODE..abilitiesList[i]..FONT_COLOR_CODE_CLOSE;
						else
							abilitiesList[i] = GREEN_FONT_COLOR_CODE..abilitiesList[i]..FONT_COLOR_CODE_CLOSE;
						end
					end
				end
						
				GameTooltip:AddLine("Abilities: "..table.concat(abilitiesList, ", "), 1.0, 1.0, 1.0);
				GameTooltip:SetHeight(GameTooltip:GetHeight() + 14);
				local width = 20 + getglobal(GameTooltip:GetName().."TextLeft"..GameTooltip:NumLines()):GetWidth();
				if ( GameTooltip:GetWidth() < width ) then
					GameTooltip:SetWidth(width);
				end
			end
		end
		
	elseif ( event == "UNIT_AURA" ) then
	
		if ( arg1 == "player" and FHH_HasTameEffect("player") ) then
			FHH_State.TamingCritter = UnitName("target");
		end
		
	elseif ( event == "UNIT_NAME_UPDATE" ) then
	
		if ( arg1 == "pet" and FHH_State.TamingCritter ) then
			FHH_CheckPetSpells(FHH_State.TamingCritter);
			FHH_State.TamingCritter = nil;
		end
		
	elseif ( event == "CRAFT_SHOW" ) then

		-- Beast Training uses the CraftFrame; we can tell it's not really a craft because it doesn't have a skill-level bar.
		local name, rank, maxRank = GetCraftDisplaySkillLine();
		if ( name ) then return; end
		
		local numCrafts = GetNumCrafts();
		if not ( numCrafts > 0 )then return; end
		
		for i=1, numCrafts do
			local craftName, craftSubSpellName = GetCraftInfo(i);
			_, _, rankNum = string.find(craftSubSpellName, "(%d)");
			local spellInfo = craftName.." "..rankNum;
			if (GFWTable.IndexOf(FHH_AbilityInfo[FHH_State.RealmPlayer], spellInfo) == 0) then
				table.insert(FHH_AbilityInfo[FHH_State.RealmPlayer], spellInfo)
			end;
		end
		table.sort(FHH_AbilityInfo[FHH_State.RealmPlayer]);

	elseif ( event == "CHAT_MSG_SYSTEM" ) then

		_, _, spellName, rankNum = string.find(arg1, "You have learned a new spell: ([%w ]+) %(Rank (%d+)%)");
		if (spellName and rankNum and spellName ~= "" and rankNum ~= "" ) then
			normalizedSpellName = string.lower(spellName);
			if (string.find(normalizedSpellName, "resist")) then
				normalizedSpellName = "resist"; -- all resistances have the same info; this lets us keep one table for them
			end
			if (FHH_RequiredLevel[normalizedSpellName] ~= nil) then
				-- only track spells we know are hunter pet spells
				table.insert(FHH_AbilityInfo[FHH_State.RealmPlayer], spellName.." "..rankNum);
				table.sort(FHH_AbilityInfo[FHH_State.RealmPlayer]);
			end
		end
		
	end

end

function FHH_ChatCommandHandler(msg)

	-- Print Help
	if ( msg == "help" ) or ( msg == "" ) then
		GFWUtils.Print("Fizzwidget Hunter's Helper:");
		GFWUtils.Print("/huntershelper /hh <command>");
		GFWUtils.Print("- "..GFWUtils.Hilite("help").." - Print this helplist.");
		GFWUtils.Print("- "..GFWUtils.Hilite("on").." | "..GFWUtils.Hilite("off").." | "..GFWUtils.Hilite("onlyhunter").." - Turn display of beast abilities in tooltips on or off, or make them only appear if you're playing a hunter.");
		GFWUtils.Print("- "..GFWUtils.Hilite("status").." - Check current settings.");
		GFWUtils.Print("- "..GFWUtils.Hilite("find <ability> <rank>").." - List where beasts with a given ability (e.g. Bite 6) can be found.");
		return;
	end
	
	if (msg == "onlyhunter") then
		FHH_Config.Tooltip = "hunter";
		GFWUtils.Print("Hunter's Helper is enabled; beast tooltips will display ability info only when playing a Hunter character.");
		return;
	end
	if (msg == "on") then
		FHH_Config.Tooltip = true;
		GFWUtils.Print("Hunter's Helper is enabled; beast tooltips will display ability info.");
		return;
	end
	if (msg == "off") then
		FHH_Config.Tooltip = nil;
		GFWUtils.Print("Hunter's Helper is disabled; no extra info added to tooltips.");
		return;
	end

	-- Check Status
	if ( msg == "status" ) then
		if ( FHH_Config.Tooltip == "hunter" ) then
			GFWUtils.Print("Hunter's Helper is enabled; beast tooltips will display ability info only when playing as a Hunter.");
		elseif ( FHH_Config.Tooltip ) then
			GFWUtils.Print("Hunter's Helper is enabled; beast tooltips will display ability info.");
		else
			GFWUtils.Print("Hunter's Helper is disabled; no extra info added to tooltips.");
		end
		return;
	end
		
	local _, _, cmd, spellName, rankNum = string.find(msg, "(find%w*) ([%l ]+) *(%d*)");
	if (cmd == "find" or cmd == "findall") then
		if (spellName == nil or spellName == "") then
			GFWUtils.Print("Usage: "..GFWUtils.Hilite("/hh find <ability> <rank>"));
			return;
		end
		
		spellName = string.lower(spellName);
		spellName = string.gsub(spellName, "^%s+", "");
		spellName = string.gsub(spellName, "%s+$", "");
		local capWords = function(first, rest) return string.upper(first)..string.lower(rest); end
		local niceSpellName = string.gsub(spellName, "(%l)([%w_']*)", capWords);
		if (string.find(spellName, "resist")) then
			spellName = "resist"; -- all resistances have the same info; this lets us keep one table for them
			niceSpellName = "Arcane/Fire/Frost/Nature/Shadow Resistance"
		end
		local spellInfo = FHH_SpellInfo[spellName];
		if (spellInfo == nil or spellInfo == {}) then
			GFWUtils.Print(GFWUtils.Hilite(niceSpellName).." is not a known beast ability.");
			return;
		end
		local rankTable = FHH_RequiredLevel[spellName];
		if (rankTable == nil or table.getn(rankTable) == 0) then
			GFWUtils.Print(GFWUtils.Red("Hunter's Helper error:").." found "..GFWUtils.Hilite(niceSpellname).." but can't find rank info. Please report to gazmik@fizzwidget.com");
			return;
		end
		
		rankNum = tonumber(rankNum);
		if (rankNum ~= nil) then
			if not (rankTable[rankNum]) then
				GFWUtils.Print(GFWUtils.Hilite(niceSpellName).." is not known to have a rank"..GFWUtils.Hilite(rankNum)..".");
				return;
			end
			
			-- report minimum pet level for ability
			local minLevel = rankTable[rankNum];
			local petLevel = 60;
			if (UnitExists("pet")) then
				petLevel = tonumber(UnitLevel("pet"));
			end
			local minLevel = FHH_RequiredLevel[spellName][rankNum];
			if (minLevel == nil) then
				GFWUtils.Print(GFWUtils.Red("Hunter's Helper error:").." can't find required level for "..GFWUtils.Hilite(niceSpellname.." "..rankNum)..". Please report to gazmik@fizzwidget.com");
			else
				if (petLevel >= minLevel) then
					GFWUtils.Print(GFWUtils.Hilite(niceSpellName.." "..rankNum).." requires pet level "..GFWUtils.Hilite(minLevel)..".");
				else
					GFWUtils.Print(GFWUtils.Hilite(niceSpellName.." "..rankNum).." requires pet level "..GFWUtils.Red(minLevel)..".");
				end
			end
		else
			GFWUtils.Print(GFWUtils.Hilite(niceSpellName).." has "..GFWUtils.Hilite(table.getn(rankTable)).." ranks.");
			if (type(spellInfo) ~= "string") then
				GFWUtils.Print("Type "..GFWUtils.Hilite("/hh find "..spellName).." and a number to get info about that rank.");
			end
		end
	
		-- report available creature families
		local families = FHH_LearnableBy[spellName];
		if (families ~= nil) then
			if (type(families) == "string") then
				GFWUtils.Print(GFWUtils.Hilite(niceSpellName).." is learnable by "..GFWUtils.Hilite(families)..".");
			else
				local listText = table.concat(families, ", ");
				GFWUtils.Print(GFWUtils.Hilite(niceSpellName).." is learnable by: "..GFWUtils.Hilite(listText)..".");
			end
		end

		-- case 1: first levels of Growl are innate
		if (spellName == "growl" and rankNum ~= nil and rankNum <= 2) then
			GFWUtils.Print("You should already know "..GFWUtils.Hilite(niceSpellName.." "..rankNum).." if you've learned Beast Training.");
			return;
		end
		
		-- case 2: spells taught by trainers, for which rank doesn't matter
		if (type(spellInfo) == "string") then
			local spellSummary = niceSpellName;
			if (rankNum ~= nil) then
				spellSummary = spellSummary.." "..rankNum;
			end
			GFWUtils.Print(GFWUtils.Hilite(spellSummary).." is learned from "..spellInfo..".");
			return;
		end
		
		if (rankNum == nil) then return; end
		
		--case 3: lookup by spell and rank, report by zone (sanity check first)
		local spellRankInfo = spellInfo[rankNum];
		local maxZones = MAX_REPORTED_ZONES;
		if (cmd == "findall") then
			maxZones = 100; -- arbitrarily high so we find everything.
		end
		if (spellRankInfo == nil or spellRankInfo == {} ) then
			GFWUtils.Print(GFWUtils.Red("Hunter's Helper error:").." expected but can't find rank-location info for "..GFWUtils.Hilite(niceSpellname.." "..rankNum)..". Please report to gazmik@fizzwidget.com.");
			return;
		end
		GFWUtils.Print(GFWUtils.Hilite(niceSpellName.." "..rankNum).." can be learned from:");
		local numReportedZones = 0;
		local zoneName = GFWZones.NormalizeZoneName(GetRealZoneText());
		if (spellRankInfo[zoneName] ~= nil and table.getn(spellRankInfo[zoneName]) > 0) then
			GFWUtils.Print(zoneName..": "..GFWUtils.Hilite(FHH_CreatureListString(spellRankInfo[zoneName])));
			numReportedZones = numReportedZones + 1;
		end
		local zones = GFWZones.ConnectionsForZone(zoneName);
		for numSteps=1, table.getn(zones) do
			for zoneIndex=1, table.getn(zones[numSteps]) do
				local zoneMobList = spellRankInfo[zones[numSteps][zoneIndex]];
				if (zoneMobList ~= nil and table.getn(zoneMobList) > 0) then
					GFWUtils.Print(GFWUtils.Hilite(zones[numSteps][zoneIndex]..": ")..FHH_CreatureListString(zoneMobList));
					numReportedZones = numReportedZones + 1;
					if (numReportedZones >= maxZones) then return; end
				end
			end
		end
		
		if (numReportedZones == 0) then
			-- if we get here, we think we know about the spell but can't find beasts with it in our table. this shouldn't happen.
			GFWUtils.Print(GFWUtils.Red("Hunter's Helper error:").." got spell info for "..GFWUtils.Hilite(niceSpellname.." "..rankNum).." but no zone info. Please report to gazmik@fizzwidget.com.");
		end
		return;
	end
	
	-- if we got all the way to here, we got invalid input.
	FHH_ChatCommandHandler("help");
	
end

function FHH_CreatureListString(critterList)
	local listString = ""
	for _, name in critterList do
		local info = FHH_BeastLevels[name];
		if (info == nil) then
			listString = listString..", ";
		else
			listString = listString .. name .. " ";
			local myLevel = UnitLevel("player");
			local minLevel = info.min;
			local maxLevel = info.max;
			if (info.min > UnitLevel("player")) then
				minLevel = RED_FONT_COLOR_CODE..info.min..FONT_COLOR_CODE_CLOSE;
			end
			if (info.max ~= nil and info.max > UnitLevel("player")) then
				maxLevel = RED_FONT_COLOR_CODE..info.max..FONT_COLOR_CODE_CLOSE;
			end
			if (info.min == info.max or info.max == nil) then			
				listString = listString.."("..minLevel;
			else
				listString = listString.."("..minLevel.."-"..maxLevel;
			end
			if (info.type == nil) then
				listString = listString.."), ";
			else
				listString = listString.." "..info.type.."), ";
			end				
		end
	end
	listString = string.gsub(listString, ", $", "");
	return listString;
end

function FHH_HasTameEffect(unit)

	local i = 1;
	local buff;
	buff = UnitBuff(unit, i);
	while buff do
		if ( string.find(buff, "Ability_Hunter_BeastTaming") ) then
			return true;
		end
		i = i + 1;
		buff = UnitBuff(unit, i);
	end
	return false;

end

function FHH_CheckPetSpells(critter)

	local CurrentPetSpells = { };
	local i = 1;
	local spellName, spellRank = GetSpellName(i, BOOKTYPE_PET);
	while spellName do
		_, _, rankNum = string.find(spellRank, "(%d)");
		table.insert(CurrentPetSpells, spellName.." "..rankNum);
		i = i + 1;
		spellName, spellRank = GetSpellName(i, BOOKTYPE_PET);
	end
	
	if ( table.getn(CurrentPetSpells) == 0 ) then return; end
	
	if ( FHH_BeastInfo[critter] ) then
	
		-- record any spells the critter has that our built-in table doesn't know about 
		local UnknownPetSpells = { };
		for i=1, table.getn(CurrentPetSpells) do
			if ( GFWTable.IndexOf(FHH_BeastInfo[critter], CurrentPetSpells[i]) == 0 ) then
				table.insert(UnknownPetSpells, CurrentPetSpells[i]);
			end
		end
		if ( table.getn(UnknownPetSpells) > 0 ) then
			if (FHH_NewBeastInfo == nil) then
				FHH_NewBeastInfo = {};
			end
			FHH_NewBeastInfo[critter] = CurrentPetSpells; -- we want to remember the entire current spells list
			local details = "(expected "..table.concat(FHH_BeastInfo[critter], ", ").."; found "..concat(CurrentPetSpells, ", ")..").";
			GFWUtils.Print("Hunter's Helper has incorrect data on "..GFWUtils.Hilite(critter..details).." Please submit a correction to gazmik@fizzwidget.com.)");
		end
		
		-- record any spells our built-in table thinks the critter has, but the critter actually doesn't
		local WrongPetSpells = { };
		for i=1, table.getn(FHH_BeastInfo[critter]) do
			if ( GFWTable.IndexOf(CurrentPetSpells, FHH_BeastInfo[critter][i]) == 0 ) then
				table.insert(WrongPetSpells, FHH_BeastInfo[critter][i]);
			end
		end
		if ( table.getn(UnknownPetSpells) > 0 ) then
			if (FHH_BadBeastInfo == nil) then
				FHH_BadBeastInfo = {};
			end
			FHH_BadBeastInfo[critter] = WrongPetSpells;
			local details = "(expected "..table.concat(FHH_BeastInfo[critter], ", ").."; found "..concat(CurrentPetSpells, ", ")..").";
			GFWUtils.Print("Hunter's Helper has incorrect data on "..GFWUtils.Hilite(critter..details).." Please submit a correction to gazmik@fizzwidget.com.)");
		end
		
	else
	
		-- this pet is entirely new to our list
		if (FHH_NewBeastInfo == nil) then
			FHH_NewBeastInfo = {};
		end
		if (FHH_NewBeastZones == nil) then
			FHH_NewBeastZones = {};
		end
		FHH_NewBeastInfo[critter] = CurrentPetSpells;
		table.insert(FHH_NewBeastZones, GetRealZoneText());
		local details = "(found "..table.concat(CurrentPetSpells, ", ").." in "..GetRealZoneText()..").";
		GFWUtils.Print("Hunter's Helper has no data on "..GFWUtils.Hilite(critter..details).." Please submit a correction to gazmik@fizzwidget.com.)");

	end
	
end

